// Note that not all keys in this file are exposed in the UI. Some of them are not used at all, but are kept for reference.
#define DeleteBackup @"DeleteBackup"
#define RichText @"RichText"
#define ShowPageBreaks @"ShowPageBreaks"
#define AddExtensionToNewPlainTextFiles @"AddExtensionToNewPlainTextFiles"
#define WindowWidth @"WidthInChars"
#define WindowHeight @"HeightInChars"
#define PlainTextEncodingForRead @"PlainTextEncoding"
#define PlainTextEncodingForWrite @"PlainTextEncodingForWrite"
#define IgnoreRichText @"IgnoreRichText"
#define IgnoreHTML @"IgnoreHTML"
#define TabWidth @"TabWidth"
#define ForegroundLayoutToIndex @"ForegroundLayoutToIndex"
#define OpenPanelFollowsMainWindow @"OpenPanelFollowsMainWindow"
#define CheckSpellingAsYouType @"CheckSpellingWhileTyping"
#define CheckGrammarWithSpelling @"CheckGrammarWithSpelling"
#define CorrectSpellingAutomatically @"CorrectSpellingAutomatically"
#define ShowRuler @"ShowRuler"
#define SmartCopyPaste @"SmartCopyPaste"
#define SmartQuotes @"SmartQuotes"
#define SmartDashes @"SmartDashes"
#define SmartLinks @"SmartLinks"
#define DataDetectors @"DataDetectors"
#define TextReplacement @"TextReplacement"
#define SubstitutionsEnabledInRichTextOnly @"SubstitutionsEnabledInRichTextOnly"
#define UseXHTMLDocType @"UseXHTMLDocType"
#define UseTransitionalDocType @"UseTransitionalDocType"
#define UseEmbeddedCSS @"UseEmbeddedCSS"
#define UseInlineCSS @"UseInlineCSS"
#define HTMLEncoding @"HTMLEncoding"
#define PreserveWhitespace @"PreserveWhitespace"
#define AutosaveDelay @"AutosaveDelay"
#define NumberPagesWhenPrinting @"NumberPagesWhenPrinting"

// Use different convention for the key values here, to be consistent with the keys in Document
#define AuthorProperty @"author"
#define CompanyProperty @"company"
#define CopyrightProperty @"copyright"
